import re
import unittest.mock

import pytest
from redis.sentinel import Sentinel

from freiner.errors import FreinerConfigurationError
from freiner.storage.redis_sentinel import RedisSentinelStorage
from freiner.types import Host

from ._redis import (
    _test_fixed_window,
    _test_fixed_window_clear,
    _test_moving_window_clear,
    _test_moving_window_expiry,
    _test_reset,
)


@pytest.fixture
def default_host() -> Host:
    return "localhost", 26379


@pytest.fixture
def default_service_name() -> str:
    return "localhost-redis-sentinel"


@pytest.fixture
def default_host_uri(default_host: Host) -> str:
    return f"redis+sentinel://{default_host[0]}:{default_host[1]}"


@pytest.fixture
def default_client(default_host: Host) -> Sentinel:
    return Sentinel((default_host,))


@pytest.fixture(autouse=True)
def flush_default_host(default_client: Sentinel, default_service_name: str):
    default_client.master_for(default_service_name).flushall()


@pytest.fixture(params=((default_client, default_service_name),))
def storage(request) -> RedisSentinelStorage:
    return RedisSentinelStorage(
        request.getfixturevalue(request.param[0].__name__),
        request.getfixturevalue(request.param[1].__name__),
    )


def test_from_default_uri(default_host_uri: str, default_service_name: str):
    storage = RedisSentinelStorage.from_uri(default_host_uri, service_name=default_service_name)
    assert isinstance(storage, RedisSentinelStorage)
    assert isinstance(storage._sentinel, Sentinel)
    assert storage.check() is True


def test_from_default_uri_with_service_name(default_host_uri: str, default_service_name: str):
    uri = default_host_uri + "/" + default_service_name
    storage = RedisSentinelStorage.from_uri(uri)
    assert isinstance(storage, RedisSentinelStorage)
    assert isinstance(storage._sentinel, Sentinel)
    assert storage.check() is True


def test_from_default_uri_with_options(default_host_uri: str, default_service_name: str):
    with unittest.mock.patch("freiner.storage.redis_sentinel.Sentinel") as mock_sentinel:
        storage = RedisSentinelStorage.from_uri(
            default_host_uri, service_name=default_service_name, connection_timeout=1
        )
        assert isinstance(storage, RedisSentinelStorage)
        assert mock_sentinel.call_args[1]["connection_timeout"] == 1


def test_from_uri_with_password(default_host: Host, default_service_name: str):
    uri = f"redis+sentinel://:sekret@{default_host[0]}:{default_host[1]}/{default_service_name}"
    with unittest.mock.patch("freiner.storage.redis_sentinel.Sentinel") as mock_sentinel:
        storage = RedisSentinelStorage.from_uri(uri)
        assert isinstance(storage, RedisSentinelStorage)
        assert mock_sentinel.call_args[1]["password"] == "sekret"


def test_from_default_uri_with_no_service_name(default_host_uri: str):
    errmsg = re.escape("'service_name' not provided")
    with pytest.raises(FreinerConfigurationError, match=errmsg):
        RedisSentinelStorage.from_uri(default_host_uri)


def test_fixed_window(storage: RedisSentinelStorage):
    _test_fixed_window(storage)


def test_fixed_window_clear(storage: RedisSentinelStorage):
    _test_fixed_window_clear(storage)


def test_moving_window_expiry(storage: RedisSentinelStorage):
    _test_moving_window_expiry(storage)


def test_moving_window_clear(storage: RedisSentinelStorage):
    _test_moving_window_clear(storage)


def test_reset(storage: RedisSentinelStorage):
    _test_reset(storage)



import unittest
import unittest.mock
from pathlib import Path
from typing import Tuple

import pytest
import redis

from freiner.storage.redis import RedisStorage
from freiner.types import Host

from ..util import DOCKERDIR
from ._redis import (
    _test_fixed_window,
    _test_fixed_window_clear,
    _test_moving_window_clear,
    _test_moving_window_expiry,
    _test_reset,
)


ProtectedHost = Tuple[str, int, str]


@pytest.fixture
def default_host() -> Host:
    return "localhost", 7379


@pytest.fixture
def protected_host() -> ProtectedHost:
    return "localhost", 7389, "sekret"


@pytest.fixture
def unix_socket_path() -> Path:
    return DOCKERDIR / "redis" / "freiner.redis.sock"


@pytest.fixture
def default_host_uri(default_host: Host) -> str:
    return f"redis://{default_host[0]}:{default_host[1]}"


@pytest.fixture
def protected_host_uri(protected_host: ProtectedHost) -> str:
    return f"redis://:{protected_host[2]}@{protected_host[0]}:{protected_host[1]}"


@pytest.fixture
def unix_socket_host_uri(unix_socket_path: Path) -> str:
    return "unix://" + str(unix_socket_path)


@pytest.fixture
def default_client(default_host_uri: str) -> redis.Redis:
    return redis.from_url(default_host_uri)


@pytest.fixture
def protected_client(protected_host_uri: str) -> redis.Redis:
    return redis.from_url(protected_host_uri)


@pytest.fixture
def unix_socket_client(unix_socket_host_uri: str) -> redis.Redis:
    return redis.from_url(unix_socket_host_uri)


@pytest.fixture(autouse=True)
def flush_default_host(default_client: redis.Redis):
    default_client.flushall()


@pytest.fixture(autouse=True)
def flush_protected_host(protected_client: redis.Redis):
    protected_client.flushall()


@pytest.fixture(autouse=True)
def flush_unix_socket_host(unix_socket_client: redis.Redis):
    unix_socket_client.flushall()


@pytest.fixture(params=(default_client, protected_client, unix_socket_client))
def client(request) -> redis.Redis:
    return request.getfixturevalue(request.param.__name__)


def test_from_uri(default_host_uri: str):
    storage = RedisStorage.from_uri(default_host_uri)
    assert isinstance(storage, RedisStorage)
    assert isinstance(storage._client, redis.Redis)
    assert storage.check() is True


def test_from_uri_with_options(default_host_uri: str):
    with unittest.mock.patch("freiner.storage.redis.redis") as mock_redis:
        storage = RedisStorage.from_uri(default_host_uri, connection_timeout=1)
        assert isinstance(storage, RedisStorage)
        assert mock_redis.from_url.call_args[1]["connection_timeout"] == 1


def test_fixed_window(client: redis.Redis):
    storage = RedisStorage(client)
    _test_fixed_window(storage)


def test_fixed_window_clear(client: redis.Redis):
    storage = RedisStorage(client)
    _test_fixed_window_clear(storage)


def test_moving_window_expiry(client: redis.Redis):
    storage = RedisStorage(client)
    _test_moving_window_expiry(storage)


def test_moving_window_clear(client: redis.Redis):
    storage = RedisStorage(client)
    _test_moving_window_clear(storage)


def test_reset(client: redis.Redis):
    storage = RedisStorage(client)
    _test_reset(storage)



import threading
import time
from typing import List
from uuid import uuid4

import pytest

from freiner.limits import RateLimitItemPerSecond
from freiner.storage.memory import MemoryStorage
from freiner.strategies.fixed_window import FixedWindowRateLimiter
from freiner.strategies.moving_window import MovingWindowRateLimiter


@pytest.fixture
def storage() -> MemoryStorage:
    return MemoryStorage()


def test_memory_storage_fixed_window(storage: MemoryStorage):
    limiter = FixedWindowRateLimiter(storage)
    per_second = RateLimitItemPerSecond(100)

    for _ in range(1000):
        limiter.hit(per_second, uuid4().hex)

    key = uuid4().hex
    hits: List[None] = []

    def hit() -> None:
        if limiter.hit(per_second, key):
            hits.append(None)

    start = time.time()

    threads = [threading.Thread(target=hit) for _ in range(1000)]

    for thread in threads:
        thread.start()

    for thread in threads:
        thread.join()

    assert time.time() - start < 1
    assert len(hits) == 100


def test_memory_storage_moving_window(storage: MemoryStorage):
    limiter = MovingWindowRateLimiter(storage)
    per_second = RateLimitItemPerSecond(100)

    for _ in range(100):
        limiter.hit(per_second, uuid4().hex)

    key = uuid4().hex
    hits: List[None] = []

    def hit() -> None:
        if limiter.hit(per_second, key):
            hits.append(None)

    start = time.time()

    threads = [threading.Thread(target=hit) for _ in range(1000)]

    for thread in threads:
        thread.start()

    for thread in threads:
        thread.join()

    assert time.time() - start < 1
    assert len(hits) == 100



import re
import time

import pytest

from freiner.storage import MovingWindow
from freiner.strategies.fixed_window import FixedWindowRateLimiter
from freiner.strategies.moving_window import MovingWindowRateLimiter


def test_pluggable_storage_fixed_window():
    class MyStorage:
        def incr(self, key: str, expiry: int, elastic_expiry: bool = False) -> int:
            return 1

        def get(self, key: str) -> int:
            return 0

        def get_expiry(self, key: str) -> float:
            return time.time()

        def clear(self, key: str):
            pass

    storage = MyStorage()
    strategy = FixedWindowRateLimiter(storage)
    assert strategy.storage is storage

    errmsg = re.escape(
        "Moving Window rate limiting is not implemented for storage of type MyStorage"
    )
    with pytest.raises(TypeError, match=errmsg):
        # Ignore the type error here because that's exactly what we're testing for.
        MovingWindowRateLimiter(storage)  # type: ignore


def test_pluggable_storage_moving_window():
    class MyStorage:
        def acquire_entry(self, key: str, limit: int, expiry: int, no_add: bool = False) -> bool:
            return True

        def get_moving_window(self, key: str, limit: int, expiry: int) -> MovingWindow:
            return MovingWindow(time.time(), 0)

        def clear(self, key: str):
            pass

    storage = MyStorage()
    strategy = MovingWindowRateLimiter(storage)
    assert strategy.storage is storage

    errmsg = re.escape(
        "Fixed Window rate limiting is not implemented for storage of type MyStorage"
    )
    with pytest.raises(TypeError, match=errmsg):
        # Ignore the type error here because that's exactly what we're testing for.
        FixedWindowRateLimiter(storage)  # type: ignore



import re

import pytest

from freiner import limits
from freiner.util import granularity_from_string, parse, parse_many


def _test_single_rate(rl_string: str, limit: limits.RateLimitItem) -> None:
    assert parse(rl_string) == limit

    many = parse_many(rl_string)
    assert len(many) == 1
    assert many[0] == limit


@pytest.mark.parametrize("rl_string", ("1 per second", "1/SECOND", "1 / Second"))
def test_single_seconds(rl_string: str):
    _test_single_rate(rl_string, limits.RateLimitItemPerSecond(1))


@pytest.mark.parametrize("rl_string", ("1 per minute", "1/MINUTE", "1/Minute"))
def test_single_minutes(rl_string: str):
    _test_single_rate(rl_string, limits.RateLimitItemPerMinute(1))


@pytest.mark.parametrize("rl_string", ("1 per hour", "1/HOUR", "1/Hour"))
def test_single_hours(rl_string: str):
    _test_single_rate(rl_string, limits.RateLimitItemPerHour(1))


@pytest.mark.parametrize("rl_string", ("1 per day", "1/DAY", "1 / Day"))
def test_single_days(rl_string: str):
    _test_single_rate(rl_string, limits.RateLimitItemPerDay(1))


@pytest.mark.parametrize("rl_string", ("1 per month", "1/MONTH", "1 / Month"))
def test_single_months(rl_string: str):
    _test_single_rate(rl_string, limits.RateLimitItemPerMonth(1))


@pytest.mark.parametrize("rl_string", ("1 per year", "1/Year", "1 / year"))
def test_single_years(rl_string: str):
    _test_single_rate(rl_string, limits.RateLimitItemPerYear(1))


@pytest.mark.parametrize(
    ("rl_string", "expiry"),
    (
        ("2 per 6 second", 6),
        ("10 per 5 minute", 5 * 60),
        ("1 per 3 hour", 3 * 60 * 60),
        ("400 per 3 day", 3 * 24 * 60 * 60),
        ("10000 per 6 month", 6 * 30 * 24 * 60 * 60),
        ("100000 per 2 year", 2 * 12 * 30 * 24 * 60 * 60),
    ),
)
def test_multiples(rl_string: str, expiry: int):
    assert parse(rl_string).get_expiry() == expiry
    assert parse(rl_string + "s").get_expiry() == expiry
    assert parse(rl_string.replace("per", "/")).get_expiry() == expiry
    assert parse(rl_string.replace(" per ", "/")).get_expiry() == expiry
    assert parse(rl_string.replace("per", "/") + "s").get_expiry() == expiry
    assert parse(rl_string.replace(" per ", "/") + "s").get_expiry() == expiry


@pytest.mark.parametrize("sep", (",", ";", "|"))
def test_parse_two_limits(sep: str):
    def _run(limit_string: str) -> None:
        parsed = parse_many(limit_string)

        assert len(parsed) == 2

        assert parsed[0].amount == 5
        assert parsed[0].get_expiry() == 3 * 60 * 60

        assert parsed[1].amount == 2
        assert parsed[1].get_expiry() == 1

        single_parsed = parse(limit_string)
        assert single_parsed == parsed[0]

    inputs = ("5 per 3 hour", "2 per second")
    for i in range(3):
        whitespace = " " * i
        full_sep = whitespace + sep + whitespace
        inputs_string = full_sep.join(inputs)
        _run(inputs_string)


@pytest.mark.parametrize("sep", (",", ";", "|"))
def test_parse_three_limits(sep):
    def _run(limit_string: str) -> None:
        parsed = parse_many(limit_string)

        assert len(parsed) == 3

        assert parsed[0].amount == 1000
        assert parsed[0].get_expiry() == 1 * 30 * 24 * 60 * 60

        assert parsed[1].amount == 200
        assert parsed[1].get_expiry() == 4 * 24 * 60 * 60

        assert parsed[2].amount == 5
        assert parsed[2].get_expiry() == 2 * 60

        single_parsed = parse(limit_string)
        assert single_parsed == parsed[0]

    inputs = ("1000 per month", "200 per 4 days", "5 per 2 minutes")
    for i in range(3):
        whitespace = " " * i
        full_sep = whitespace + sep + whitespace
        inputs_string = full_sep.join(inputs)
        _run(inputs_string)


def test_invalid_input():
    errmsg = re.escape("Invalid rate limit string supplied.")

    with pytest.raises(TypeError, match=errmsg):
        parse(None)  # type: ignore

    with pytest.raises(TypeError, match=errmsg):
        parse_many(None)  # type: ignore


def test_invalid_string():
    errmsg = "^" + re.escape("Failed to parse rate limit string: ")

    with pytest.raises(ValueError, match=errmsg):
        parse("")

    with pytest.raises(ValueError, match=errmsg):
        parse("1 per millenium")

    with pytest.raises(ValueError, match=errmsg):
        parse_many("30 per month1 per day")

    with pytest.raises(ValueError, match=errmsg):
        parse_many("30 per month 1 per day")

    with pytest.raises(ValueError, match=errmsg):
        parse_many("1 per year; 2 per decade")

    with pytest.raises(ValueError, match=errmsg):
        parse("this won't match")


@pytest.mark.parametrize("sep", (",", ";", "|"))
def test_two_limits_with_multiple_consecutive_separators(sep: str):
    errmsg = "^" + re.escape("Failed to parse rate limit string: ")
    inputs = ("200 per day", "10 per hour")

    for i in range(3):
        for j in range(2, 5):
            whitespace = " " * i
            full_sep = whitespace + (sep * j) + whitespace
            inputs_string = full_sep.join(inputs)
            with pytest.raises(ValueError, match=errmsg):
                parse_many(inputs_string)


@pytest.mark.parametrize("sep", (",", ";", "|"))
def test_three_limits_with_multiple_consecutive_separators(sep: str):
    errmsg = "^" + re.escape("Failed to parse rate limit string: ")
    inputs = ("1200 per day", "50 per hour", "1 per minute")

    for i in range(3):
        for j in range(2, 5):
            whitespace = " " * i
            full_sep = whitespace + (sep * j) + whitespace
            inputs_string = full_sep.join(inputs)
            with pytest.raises(ValueError, match=errmsg):
                parse_many(inputs_string)


def test_unknown_granularity():
    errmsg = "^" + re.escape("No granularity matched for: ")

    with pytest.raises(ValueError, match=errmsg):
        granularity_from_string("millenium")



from freiner import limits


def test_seconds_value():
    assert limits.RateLimitItemPerSecond(1).get_expiry() == 1
    assert limits.RateLimitItemPerMinute(1).get_expiry() == 60
    assert limits.RateLimitItemPerHour(1).get_expiry() == 60 * 60
    assert limits.RateLimitItemPerDay(1).get_expiry() == 24 * 60 * 60
    assert limits.RateLimitItemPerMonth(1).get_expiry() == 30 * 24 * 60 * 60
    assert limits.RateLimitItemPerYear(1).get_expiry() == 12 * 30 * 24 * 60 * 60


def test_str_representation():
    assert str(limits.RateLimitItemPerSecond(1)) == "1 per 1 second"
    assert str(limits.RateLimitItemPerMinute(1)) == "1 per 1 minute"
    assert str(limits.RateLimitItemPerHour(1)) == "1 per 1 hour"
    assert str(limits.RateLimitItemPerDay(1)) == "1 per 1 day"
    assert str(limits.RateLimitItemPerMonth(1)) == "1 per 1 month"
    assert str(limits.RateLimitItemPerYear(1)) == "1 per 1 year"


def test_repr_representation():
    assert repr(limits.RateLimitItemPerSecond(2)) == "RateLimitItemPerSecond<2 per 1 second>"
    assert repr(limits.RateLimitItemPerMinute(3)) == "RateLimitItemPerMinute<3 per 1 minute>"
    assert repr(limits.RateLimitItemPerHour(4)) == "RateLimitItemPerHour<4 per 1 hour>"
    assert repr(limits.RateLimitItemPerDay(5)) == "RateLimitItemPerDay<5 per 1 day>"
    assert repr(limits.RateLimitItemPerMonth(6)) == "RateLimitItemPerMonth<6 per 1 month>"
    assert repr(limits.RateLimitItemPerYear(7)) == "RateLimitItemPerYear<7 per 1 year>"


def test_comparison():
    assert limits.RateLimitItemPerSecond(1) < limits.RateLimitItemPerMinute(1)
    assert limits.RateLimitItemPerMinute(1) < limits.RateLimitItemPerHour(1)
    assert limits.RateLimitItemPerHour(1) < limits.RateLimitItemPerDay(1)
    assert limits.RateLimitItemPerDay(1) < limits.RateLimitItemPerMonth(1)
    assert limits.RateLimitItemPerMonth(1) < limits.RateLimitItemPerYear(1)


def test_no_granularity():
    def _define_class() -> None:
        class NoGranularity(limits.RateLimitItem):
            pass

    known_granularities_before = list(limits.GRANULARITIES.keys())
    _define_class()
    known_granularities_after = list(limits.GRANULARITIES.keys())

    assert known_granularities_after == known_granularities_before



from typing import Any

from freiner.limits import RateLimitItem
from freiner.storage import MovingWindowStorage

from . import WindowStats


class MovingWindowRateLimiter:
    """
    Reference: :ref:`moving-window`
    """

    def __init__(self, storage: MovingWindowStorage) -> None:
        if not isinstance(storage, MovingWindowStorage):
            msg = f"Moving Window rate limiting is not implemented for storage of type {storage.__class__.__name__}"
            raise TypeError(msg)

        self.storage: MovingWindowStorage = storage

    def hit(self, item: RateLimitItem, *identifiers: Any) -> bool:
        """
        Creates a hit on the rate limit and returns ``True`` if successful.

        :param item: A :class:`freiner.limits.RateLimitItem` instance.
        :param identifiers: A variable list of stringable objects to uniquely identify the limit.
        :return: ``True`` if the request was successful, or ``False`` if the rate limit had been exceeded.
        """

        return self.storage.acquire_entry(
            item.key_for(*identifiers), item.amount, item.get_expiry()
        )

    def test(self, item: RateLimitItem, *identifiers: Any) -> bool:
        """
        Checks the rate limit and returns ``True`` if it is not currently exceeded.

        :param item: A :class:`freiner.limits.RateLimitItem` instance.
        :param identifiers: A variable list of stringable objects to uniquely identify the limit.
        :return: ``True`` if the rate limit has not yet been exceeded, or ``False`` if it has.
        """

        _, acquired_count = self.storage.get_moving_window(
            item.key_for(*identifiers), item.amount, item.get_expiry()
        )

        return acquired_count < item.amount

    def get_window_stats(self, item: RateLimitItem, *identifiers: Any) -> WindowStats:
        """
        Returns the number of requests remaining within this limit.

        :param item: a :class:`freiner.limits.RateLimitItem` instance
        :param identifiers: A variable list of stringable objects to uniquely identify the limit.
        :return: tuple (reset time (float), remaining (int))
        """

        window_start, window_items = self.storage.get_moving_window(
            item.key_for(*identifiers), item.amount, item.get_expiry()
        )
        reset = window_start + item.get_expiry()
        return WindowStats(reset, item.amount - window_items)

    def clear(self, item: RateLimitItem, *identifiers: Any) -> None:
        """
        Resets the request counter for a given limit to zero.

        :param item: a :class:`freiner.limits.RateLimitItem` instance
        :param identifiers: A variable list of stringable objects to uniquely identify the limit.
        """

        self.storage.clear(item.key_for(*identifiers))


__all__ = [
    "MovingWindowRateLimiter",
]



import time
from typing import Any, List, Union
from urllib.parse import urlparse

import pymemcache

from freiner.errors import FreinerConfigurationError
from freiner.types import Host


MemcachedClient = Union[pymemcache.Client, pymemcache.PooledClient, pymemcache.HashClient]


class MemcachedStorage:
    """
    Rate limit storage with memcached as backend.

    Depends on the `pymemcache` library.
    """

    MAX_CAS_RETRIES = 10

    def __init__(self, client: MemcachedClient) -> None:
        self._client: MemcachedClient = client

    @classmethod
    def from_uri(cls, uri: str, **options: Any) -> "MemcachedStorage":
        """
        :param uri: URI of the form `memcached://host:port,host:port`or `memcached:///run/path/to/sock`.
        :param options: All remaining keyword arguments are passed directly to the constructor
                        of :class:`pymemcache.client.base.Client`.
        :raises FreinerConfigurationError: When no hosts could be parsed from the supplied URI.
        """

        parsed_uri = urlparse(uri)
        hosts: List[Union[Host, str]] = []
        for loc in parsed_uri.netloc.strip().split(","):
            if not loc:
                continue

            host, port = loc.split(":")
            hosts.append((host, int(port)))
        else:
            # filesystem path to UDS
            if parsed_uri.path and not parsed_uri.netloc and not parsed_uri.port:
                hosts = [parsed_uri.path]

        if not hosts:
            raise FreinerConfigurationError(f"No Memcached hosts parsed from URI: {uri}")

        if len(hosts) > 1:
            client = pymemcache.HashClient(hosts, **options)
        else:
            client = pymemcache.Client(*hosts, **options)
        return cls(client)

    def get(self, key: str) -> int:
        """
        Retrieve the current request count for the given rate limit key.

        :param key: The key to get the counter value for.
        """

        return int(self._client.get(key) or 0)

    def clear(self, key: str) -> None:
        """
        Resets the rate limit for the given key.

        :param key: The key to clear rate limits for.
        """
        self._client.delete(key)

    def incr(self, key: str, expiry: int, elastic_expiry: bool = False) -> int:
        """
        Increments the counter for the given rate limit key.

        :param key: The key to increment.
        :param expiry: Amount in seconds for the key to expire in.
        :param elastic_expiry: Whether to keep extending the rate limit window every hit.
        :return: The number of hits currently on the rate limit for the given key.
        """

        if self._client.add(key, 1, expiry, noreply=False):
            self._set_expiry(key, expiry)
            return 1

        if not elastic_expiry:
            return self._client.incr(key, 1) or 1

        # TODO: There is a timing issue here.
        # This code makes the assumption that because client.add() failed, the key must exist.
        # That isn't necessarily true. It can expire between us calling client.add() and us
        # calling client.gets(). If that happens, 'cas' will be None. If we pass cas=None to
        # client.cas(), it gets very unhappy.
        # This issue shows up occasionally in the test suite, both locally and on Github Actions.
        # If it shows up in testing, it absolutely will show up in the real world.
        # I believe the solution will be to "restart" the logic flow if 'cas is None'. However,
        # that will require rewriting the method so that that can be achieved without recursion,
        # and without the code looking like a nightmare.
        value, cas = self._client.gets(key)
        retry = 0

        while (
            not self._client.cas(key, int(value or 0) + 1, cas, expiry)
            and retry < self.MAX_CAS_RETRIES
        ):
            value, cas = self._client.gets(key)
            retry += 1

        self._set_expiry(key, expiry)
        return int(value or 0) + 1

    def _set_expiry(self, key: str, expiry: int):
        self._client.set(key + "/expires", expiry + time.time(), expire=expiry, noreply=False)

    def get_expiry(self, key: str) -> float:
        """
        Retrieve the expected expiry time for the given rate limit key.

        :param key: The key to get the expiry time for.
        :return: The time at which the current rate limit for the given key ends.
        """

        return float(self._client.get(key + "/expires") or time.time())

    def check(self) -> bool:
        """
        Check if the connection to the storage backend is healthy.
        """

        try:
            self._client.get("freiner-check")
            return True
        except:  # noqa
            return False


__all__ = [
    "MemcachedClient",
    "MemcachedStorage",
]



from typing import Any

from freiner.limits import RateLimitItem

from .fixed_window import FixedWindowRateLimiter


class FixedWindowElasticExpiryRateLimiter(FixedWindowRateLimiter):
    """
    Reference: :ref:`fixed-window-elastic`
    """

    def hit(self, item: RateLimitItem, *identifiers: Any) -> bool:
        """
        Creates a hit on the rate limit and returns ``True`` if successful.

        :param item: A :class:`freiner.limits.RateLimitItem` instance.
        :param identifiers: A variable list of stringable objects to uniquely identify the limit.
        :return: ``True`` if the request was successful, or ``False`` if the rate limit had been exceeded.
        """

        counter = self.storage.incr(
            item.key_for(*identifiers), item.get_expiry(), elastic_expiry=True
        )
        return counter <= item.amount


__all__ = [
    "FixedWindowElasticExpiryRateLimiter",
]



import re
import time
import unittest.mock
from pathlib import Path
from typing import Tuple

import pymemcache
import pytest

from freiner import RateLimitItemPerMinute, RateLimitItemPerSecond
from freiner.errors import FreinerConfigurationError
from freiner.storage.memcached import MemcachedStorage
from freiner.strategies.fixed_window import FixedWindowRateLimiter
from freiner.strategies.fixed_window_elastic import FixedWindowElasticExpiryRateLimiter
from freiner.types import Host

from ..util import DOCKERDIR, fixed_start


@pytest.fixture
def default_host() -> Host:
    return "localhost", 22122


@pytest.fixture
def secondary_host() -> Host:
    return "localhost", 22123


@pytest.fixture
def hash_hosts(default_host: Host, secondary_host: Host) -> Tuple[Host, ...]:
    return default_host, secondary_host


@pytest.fixture
def unix_socket_path() -> Path:
    return DOCKERDIR / "memcached" / "freiner.memcached.sock"


@pytest.fixture
def unix_socket_host_uri(unix_socket_path: Path) -> str:
    return "unix://" + str(unix_socket_path)


@pytest.fixture
def default_client(default_host: Host) -> pymemcache.Client:
    return pymemcache.Client(default_host)


@pytest.fixture
def hash_client(hash_hosts: Tuple[Host, ...]) -> pymemcache.HashClient:
    return pymemcache.HashClient(hash_hosts)


@pytest.fixture
def unix_socket_client(unix_socket_host_uri: str) -> pymemcache.Client:
    return pymemcache.Client(unix_socket_host_uri)


@pytest.fixture
def flush_default_host(default_client: pymemcache.Client):
    default_client.flush_all()


@pytest.fixture
def flush_hash_host(hash_client: pymemcache.HashClient):
    hash_client.flush_all()


@pytest.fixture
def flush_unix_socket_host(unix_socket_client: pymemcache.Client):
    unix_socket_client.flush_all()


@pytest.mark.usefixtures("flush_default_host")
def test_from_plain_uri(default_host: Host):
    plain_uri = f"memcached://{default_host[0]}:{default_host[1]}"
    plain_storage = MemcachedStorage.from_uri(plain_uri)
    assert isinstance(plain_storage, MemcachedStorage)
    assert isinstance(plain_storage._client, pymemcache.Client)
    assert plain_storage.check() is True


def test_from_plain_uri_with_options(default_host: Host):
    plain_uri = f"memcached://{default_host[0]}:{default_host[1]}"
    with unittest.mock.patch("freiner.storage.memcached.pymemcache") as mock_pymemcache:
        MemcachedStorage.from_uri(plain_uri, connect_timeout=1)
        assert mock_pymemcache.Client.call_args[1]["connect_timeout"] == 1


@pytest.mark.usefixtures("flush_hash_host")
def test_from_hash_uri(hash_hosts: Tuple[Host, ...]):
    uri_hosts = []
    for host in hash_hosts:
        uri_hosts.append(f"{host[0]}:{host[1]}")

    hash_uri = f"memcached://{','.join(uri_hosts)}"
    hash_storage = MemcachedStorage.from_uri(hash_uri)
    assert isinstance(hash_storage, MemcachedStorage)
    assert isinstance(hash_storage._client, pymemcache.HashClient)
    assert hash_storage.check() is True


@pytest.mark.usefixtures("flush_unix_socket_host")
def test_from_unix_socket_uri(unix_socket_host_uri: str):
    storage = MemcachedStorage.from_uri(unix_socket_host_uri)
    assert isinstance(storage, MemcachedStorage)
    assert isinstance(storage._client, pymemcache.Client)
    assert storage.check() is True


def test_uri_with_no_hosts():
    uri = "memcached://"
    errmsg = re.escape("No Memcached hosts parsed from URI: memcached://")
    with pytest.raises(FreinerConfigurationError, match=errmsg):
        MemcachedStorage.from_uri(uri)


@pytest.mark.usefixtures("flush_default_host")
@fixed_start
def test_fixed_window(default_client: pymemcache.Client):
    storage = MemcachedStorage(default_client)
    assert storage.check() is True

    limiter = FixedWindowRateLimiter(storage)
    per_min = RateLimitItemPerSecond(10)

    start = time.time()
    count = 0
    while time.time() - start < 0.5 and count < 10:
        assert limiter.hit(per_min) is True
        count += 1
    assert limiter.hit(per_min) is False

    while time.time() - start <= 1:
        time.sleep(0.1)
    assert limiter.hit(per_min) is True


@pytest.mark.usefixtures("flush_hash_host")
@fixed_start
def test_fixed_window_cluster(hash_client: pymemcache.HashClient):
    storage = MemcachedStorage(hash_client)
    assert storage.check() is True

    limiter = FixedWindowRateLimiter(storage)
    per_min = RateLimitItemPerSecond(10)

    start = time.time()
    count = 0
    while time.time() - start < 0.5 and count < 10:
        assert limiter.hit(per_min) is True
        count += 1
    assert limiter.hit(per_min) is False

    while time.time() - start <= 1:
        time.sleep(0.1)
    assert limiter.hit(per_min) is True


@pytest.mark.usefixtures("flush_default_host")
@fixed_start
def test_fixed_window_with_elastic_expiry(default_client: pymemcache.Client):
    storage = MemcachedStorage(default_client)
    assert storage.check() is True

    limiter = FixedWindowElasticExpiryRateLimiter(storage)
    per_sec = RateLimitItemPerSecond(2, 2)

    assert limiter.hit(per_sec) is True

    time.sleep(1)
    assert limiter.hit(per_sec) is True
    assert limiter.test(per_sec) is False

    time.sleep(1)
    assert limiter.test(per_sec) is False

    time.sleep(1)
    assert limiter.test(per_sec) is True


@pytest.mark.usefixtures("flush_hash_host")
@fixed_start
def test_fixed_window_with_elastic_expiry_cluster(hash_client: pymemcache.HashClient):
    storage = MemcachedStorage(hash_client)
    assert storage.check() is True

    limiter = FixedWindowElasticExpiryRateLimiter(storage)
    per_sec = RateLimitItemPerSecond(2, 2)

    assert limiter.hit(per_sec) is True

    time.sleep(1)
    assert limiter.hit(per_sec) is True
    assert limiter.test(per_sec) is False

    time.sleep(1)
    assert limiter.test(per_sec) is False

    time.sleep(1)
    assert limiter.test(per_sec) is True


@pytest.mark.usefixtures("flush_default_host")
def test_clear(default_client: pymemcache.Client):
    storage = MemcachedStorage(default_client)
    assert storage.check() is True

    limiter = FixedWindowRateLimiter(storage)
    per_min = RateLimitItemPerMinute(1)

    assert limiter.hit(per_min) is True
    assert limiter.hit(per_min) is False

    limiter.clear(per_min)
    assert limiter.hit(per_min) is True



import time

from freiner.limits import RateLimitItemPerMinute, RateLimitItemPerSecond
from freiner.storage.redis import RedisStorage
from freiner.strategies.fixed_window import FixedWindowRateLimiter
from freiner.strategies.moving_window import MovingWindowRateLimiter


# TODO: Rewrite this to use freezegun. Will require fakeredis.
def _test_fixed_window(storage: RedisStorage) -> None:
    limiter = FixedWindowRateLimiter(storage)
    per_second = RateLimitItemPerSecond(10)

    start = time.time()
    count = 0
    while time.time() - start < 0.5 and count < 10:
        assert limiter.hit(per_second) is True
        count += 1
    assert limiter.hit(per_second) is False

    while time.time() - start <= 1:
        time.sleep(0.1)

    for _ in range(10):
        assert limiter.hit(per_second) is True


def _test_fixed_window_clear(storage: RedisStorage) -> None:
    limiter = FixedWindowRateLimiter(storage)
    per_min = RateLimitItemPerMinute(1)

    assert limiter.hit(per_min) is True
    assert limiter.hit(per_min) is False

    limiter.clear(per_min)
    assert limiter.hit(per_min) is True


# TODO: Rewrite this to use freezegun. Will require fakeredis.
def _test_moving_window_expiry(storage: RedisStorage) -> None:
    limiter = MovingWindowRateLimiter(storage)
    limit = RateLimitItemPerSecond(2)

    assert limiter.hit(limit) is True

    time.sleep(0.9)
    assert limiter.hit(limit) is True
    assert limiter.hit(limit) is False

    time.sleep(0.1)
    assert limiter.hit(limit) is True

    last = time.time()
    while time.time() - last <= 1:
        time.sleep(0.05)
    assert storage._client.keys(limit.namespace + "/*") == []


def _test_moving_window_clear(storage: RedisStorage) -> None:
    limiter = MovingWindowRateLimiter(storage)
    per_min = RateLimitItemPerMinute(1)

    assert limiter.hit(per_min) is True
    assert limiter.hit(per_min) is False

    limiter.clear(per_min)
    assert limiter.hit(per_min) is True


def _test_reset(storage) -> None:
    limiter = FixedWindowRateLimiter(storage)
    rate = RateLimitItemPerMinute(1)

    assert limiter.hit(rate) is True
    assert limiter.hit(rate) is False

    storage.reset()
    assert limiter.hit(rate) is True



from typing import Any
from urllib.parse import urlparse

from rediscluster import RedisCluster

from .redis import RedisStorage


class RedisClusterStorage(RedisStorage):
    """
    Rate limit storage with redis cluster as backend.

    Depends on `redis-py-cluster` library.
    """

    @classmethod
    def from_uri(cls, uri: str, **options: Any) -> "RedisClusterStorage":
        """
        :param uri: URI of the form `redis+cluster://[:password]@host:port,host:port`
        :param options: All remaining keyword arguments are passed directly to the constructor
                        of :class:`rediscluster.RedisCluster`.
        """

        parsed_uri = urlparse(uri)
        cluster_hosts = []
        for loc in parsed_uri.netloc.split(","):
            host, port = loc.split(":")
            cluster_hosts.append({"host": host, "port": int(port)})

        options.setdefault("max_connections", 1000)
        options["startup_nodes"] = cluster_hosts

        client = RedisCluster(**options)
        return cls(client)

    def reset(self) -> None:
        """
        Redis Clusters are sharded and deleting across shards
        can't be done atomically. Because of this, this reset loops over all
        keys that are prefixed with 'LIMITER' and calls delete on them, one at
        a time.

        .. warning::
         This operation was not tested with extremely large data sets.
         On a large production based system, care should be taken with its
         usage as it could be slow on very large data sets.
        """

        keys = self._client.keys("LIMITER*")
        for key in keys:
            self._client.delete(key.decode("utf-8"))


__all__ = [
    "RedisClusterStorage",
]



from typing import Any, Optional
from urllib.parse import urlparse

from redis import Redis
from redis.sentinel import Sentinel

from freiner.errors import FreinerConfigurationError

from .redis import RedisStorage


class RedisSentinelStorage(RedisStorage):
    """
    Rate limit storage with redis sentinel as backend.

    Depends on `redis` library.
    """

    def __init__(self, sentinel: Sentinel, service_name: str) -> None:
        self._sentinel: Sentinel = sentinel
        self._service_name: str = service_name

        self._sentinel_master: Redis = self._sentinel.master_for(self._service_name)
        self._sentinel_slave: Redis = self._sentinel.slave_for(self._service_name)

        super().__init__(self._sentinel_master)

    @classmethod
    def from_uri(
        cls, uri: str, service_name: Optional[str] = None, **options: Any
    ) -> "RedisSentinelStorage":
        """
        :param uri: URI of the form `redis+sentinel://host:port,host:port/service_name`
        :param service_name: Sentinel service name (if not provided in `uri`).
        :param options: All remaining keyword arguments are passed directly to the constructor
                        of :class:`redis.sentinel.Sentinel`.
        :raises FreinerConfigurationError: When no service name is provided.
        """

        parsed_uri = urlparse(uri)
        sentinel_configuration = []

        password = None
        if parsed_uri.password:
            password = parsed_uri.password

        for loc in parsed_uri.netloc[parsed_uri.netloc.find("@") + 1 :].split(","):
            host, port = loc.split(":")
            sentinel_configuration.append((host, int(port)))

        service_name = parsed_uri.path.replace("/", "") if parsed_uri.path else service_name
        if service_name is None:
            raise FreinerConfigurationError("'service_name' not provided")

        options.setdefault("socket_timeout", 0.2)

        sentinel = Sentinel(sentinel_configuration, password=password, **options)
        return cls(sentinel, service_name)

    def get(self, key: str) -> int:
        """
        Retrieve the current request count for the given rate limit key.

        :param key: The key to get the counter value for.
        """

        return self._get(key, self._sentinel_slave)

    def get_expiry(self, key: str) -> float:
        """
        Retrieve the expected expiry time for the given rate limit key.

        :param key: The key to get the expiry time for.
        :return: The time at which the current rate limit for the given key ends.
        """

        return self._get_expiry(key, self._sentinel_slave)

    def check(self) -> bool:
        """
        Check if the connection to the storage backend is healthy.
        """

        return self._check(self._sentinel_slave)


__all__ = [
    "RedisSentinelStorage",
]



from typing import NamedTuple

from typing_extensions import Protocol, runtime_checkable


@runtime_checkable
class FixedWindowStorage(Protocol):
    def incr(self, key: str, expiry: int, elastic_expiry: bool = False) -> int:
        """
        Increments the counter for the given rate limit key.

        # noqa: DAR202

        :param key: The key to increment.
        :param expiry: Amount in seconds for the key to expire in.
        :param elastic_expiry: Whether to keep extending the rate limit window every hit.
        :return: The number of hits currently on the rate limit for the given key.
        """

    def get(self, key: str) -> int:
        """
        Retrieve the current request count for the given rate limit key.

        :param key: The key to get the counter value for.
        """

    def get_expiry(self, key: str) -> float:
        """
        Retrieve the expected expiry time for the given rate limit key.

        # noqa: DAR202

        :param key: The key to get the expiry time for.
        :return: The time at which the current rate limit for the given key ends.
        """

    def clear(self, key: str) -> None:
        """
        Resets the rate limit for the given key.

        :param key: The key to clear rate limits for.
        """


class MovingWindow(NamedTuple):
    start_time: float
    acquired_count: int


@runtime_checkable
class MovingWindowStorage(Protocol):
    def acquire_entry(self, key: str, limit: int, expiry: int, no_add: bool = False) -> bool:
        """
        :param key: The rate limit key to acquire an entry in.
        :param limit: The total amount of entries allowed before hitting the rate limit.
        :param expiry: Amount in seconds for the acquired entry to expire in.
        :param no_add: If False, an entry is not actually acquired but instead serves as a 'check'.
        """

    def get_moving_window(self, key: str, limit: int, expiry: int) -> MovingWindow:
        """
        Retrieves the starting point and the number of entries in the moving window.

        # noqa: DAR202

        :param key: The rate limit key to retrieve statistics about.
        :param limit: The total amount of entries allowed before hitting the rate limit.
        :param expiry: Amount in seconds for the acquired entry to expire in.
        :return: (start of window, number of acquired entries)
        """

    def clear(self, key: str) -> None:
        """
        Resets the rate limit for the given key.

        :param key: The key to clear rate limits for.
        """


__all__ = [
    "FixedWindowStorage",
    "MovingWindow",
    "MovingWindowStorage",
]



from freiner import limits


class FakeLimit(limits.RateLimitItem):
    granularity = (1, "fake")


def test_key_all_strings_default_namespace():
    item = FakeLimit(1, 1)
    assert item.key_for("a", "b", "c") == "LIMITER/a/b/c/1/1/fake"


def test_key_with_none_default_namespace():
    item = FakeLimit(1, 1)
    assert item.key_for("a", None, None) == "LIMITER/a/None/None/1/1/fake"


def test_key_with_int_default_namespace():
    item = FakeLimit(1, 1)
    assert item.key_for("a", 1) == "LIMITER/a/1/1/1/fake"


def test_key_with_mixed_string_types_default_namespace():
    item = FakeLimit(1, 1)
    assert item.key_for(b"a", "b") == "LIMITER/a/b/1/1/fake"


def test_key_with_object_default_namespace():
    item = FakeLimit(1, 1)
    assert item.key_for({"a": 1, "b": 2}) == "LIMITER/{'a': 1, 'b': 2}/1/1/fake"



from functools import total_ordering
from typing import Any, Dict, Type, cast


def safe_string(value: Any) -> str:
    """
    Consistently converts a value to a string.

    :param value: The value to stringify.
    """

    if isinstance(value, bytes):
        return value.decode()
    return str(value)


TIME_TYPES = {
    "year": (60 * 60 * 24 * 30 * 12, "year"),
    "month": (60 * 60 * 24 * 30, "month"),
    "day": (60 * 60 * 24, "day"),
    "hour": (60 * 60, "hour"),
    "minute": (60, "minute"),
    "second": (1, "second"),
}

GRANULARITIES: Dict[str, Type["RateLimitItem"]] = {}


class RateLimitItemMeta(type):
    def __new__(cls, name, parents, dct):
        granularity = cast(
            Type["RateLimitItem"], super(RateLimitItemMeta, cls).__new__(cls, name, parents, dct)
        )
        if "granularity" in dct:
            GRANULARITIES[dct["granularity"][1]] = granularity
        return granularity


@total_ordering
class RateLimitItem(metaclass=RateLimitItemMeta):
    """
    defines a Rate limited resource which contains the characteristic
    namespace, amount and granularity multiples of the rate limiting window.

    :param amount: The amount of hits that can be sustained within the given period (e.g. X in 'x per y seconds').
    :param multiples: Multiple of the 'per' granularity (e.g. Y in 'x per y seconds').
    :param namespace: An arbitrary namespace for this rate limit.
    """

    granularity = (0, "")

    def __init__(self, amount: int, multiples: int = 1, namespace: str = "LIMITER") -> None:
        self.namespace = namespace
        self.amount = int(amount)
        self.multiples = int(multiples or 1)

    @classmethod
    def check_granularity_string(cls, granularity_string: str) -> bool:
        """
        Check if this class matches a granularity string of type 'N per hour' etc.

        :param granularity_string: The granularity to match against.
        :return: Does the supplied granularity identifier match this class' granularity?
        """

        return granularity_string.lower() in cls.granularity[1:]

    def get_expiry(self) -> int:
        """
        :return: The length of the expiry window in seconds.
        """

        return self.granularity[0] * self.multiples

    def key_for(self, *identifiers: Any) -> str:
        """
        :param identifiers: A list of arbitrary strings to append to the key.
        :return: A string key identifying this resource with each identifier appended with a '/' delimiter.
        """

        identifier_strings = [safe_string(k) for k in identifiers]
        limit_strings = [safe_string(self.amount), safe_string(self.multiples), self.granularity[1]]

        remainder = "/".join(identifier_strings + limit_strings)
        return "{}/{}".format(self.namespace, remainder)

    def __eq__(self, other) -> bool:
        return (
            isinstance(other, RateLimitItem)
            and self.amount == other.amount
            and self.granularity == other.granularity
        )

    def __str__(self) -> str:
        return "{} per {} {}".format(self.amount, self.multiples, self.granularity[1])

    def __repr__(self) -> str:
        return "{}<{}>".format(self.__class__.__name__, str(self))

    def __lt__(self, other) -> bool:
        return isinstance(other, RateLimitItem) and self.granularity[0] < other.granularity[0]


class RateLimitItemPerYear(RateLimitItem):
    """
    Per year rate limited resource.
    """

    granularity = TIME_TYPES["year"]


class RateLimitItemPerMonth(RateLimitItem):
    """
    Per month rate limited resource.
    """

    granularity = TIME_TYPES["month"]


class RateLimitItemPerDay(RateLimitItem):
    """
    Per day rate limited resource.
    """

    granularity = TIME_TYPES["day"]


class RateLimitItemPerHour(RateLimitItem):
    """
    Per hour rate limited resource.
    """

    granularity = TIME_TYPES["hour"]


class RateLimitItemPerMinute(RateLimitItem):
    """
    Per minute rate limited resource.
    """

    granularity = TIME_TYPES["minute"]


class RateLimitItemPerSecond(RateLimitItem):
    """
    Per second rate limited resource.
    """

    granularity = TIME_TYPES["second"]



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sphinx_rtd_theme
import sys

sys.path.insert(0, os.path.abspath("../../"))
import freiner


# -- Project information -----------------------------------------------------

project = "Freiner"
copyright = "2021, Matthew Gamble"
author = "Matthew Gamble"

# The full version, including alpha/beta/rc tags
version = release = freiner.__version__


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
    "sphinx_search.extension",
]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

pygments_style = "sphinx"


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.

html_theme = "sphinx_rtd_theme"
html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]

# -- Extensions --------------------------------------------------------------

autodoc_default_options = {
    "members": True,
    "member-order": "bysource",
    "show-inheritance": True,
}
autoclass_content = "both"
autodoc_typehints = "both"

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "redis": ("https://redis-py.readthedocs.io/en/stable/", None),
    "rediscluster": ("https://redis-py-cluster.readthedocs.io/en/stable/", None),
    "pymemcache": ("https://pymemcache.readthedocs.io/en/stable/", None),
}



import os
import sys

from invoke import task, util


in_ci = os.environ.get("CI", "false") == "true"
if in_ci:
    pty = False
else:
    pty = util.isatty(sys.stdout) and util.isatty(sys.stderr)


@task
def reformat(c):
    c.run("isort freiner tests setup.py tasks.py", pty=pty)
    c.run("black freiner tests setup.py tasks.py docs/source/conf.py", pty=pty)


@task
def lint(c):
    c.run("flake8 --show-source --statistics --max-line-length 100 freiner tests", pty=pty)
    # TODO: start using pydocstyle
    # Gotta use pydocstyle directly rather than through the flake8 config, because the flake8
    # plugin doesn't support configuring pydocstyle sensibly.

    c.run("check-manifest", pty=pty)

    bandit_args = ["bandit", "--configfile", "bandit.yaml", "-r"]
    if pty:
        bandit_args.extend(("-f", "screen"))
    if not in_ci:
        bandit_args.append("--quiet")
    bandit_args.extend(("freiner", "tests"))
    c.run(" ".join(bandit_args), pty=pty)


@task
def test(c, onefile=""):
    pytest_args = ["pytest", "--strict-config", "--cov=freiner", "--cov-report=term"]
    if in_ci:
        pytest_args.extend(("--cov-report=xml", "--strict-markers"))
    else:
        pytest_args.append("--cov-report=html")

    if onefile:
        pytest_args.append(onefile)

    c.run("docker-compose down --remove-orphans --volumes", pty=pty)
    c.run("docker-compose up -d", pty=pty)
    try:
        c.run(" ".join(pytest_args), pty=pty)
    finally:
        c.run("docker-compose down --remove-orphans --volumes", pty=pty)

        c.run("rm .docker/memcached/freiner.memcached.sock", pty=pty, warn=True)
        c.run("rm .docker/redis/freiner.redis.sock", pty=pty, warn=True)


@task
def type_check(c):
    c.run("mypy freiner tests", pty=pty)


@task
def docs(c):
    with c.cd("docs"):
        c.run("sphinx-build -M html source build -a -W")



import threading
import time
import traceback

import pymemcache
import pytest

from freiner.limits import RateLimitItemPerSecond
from freiner.storage.memcached import MemcachedStorage
from freiner.strategies.fixed_window_elastic import FixedWindowElasticExpiryRateLimiter
from freiner.strategies.moving_window import MovingWindowRateLimiter
from freiner.types import Host


@pytest.fixture
def default_host() -> Host:
    return "localhost", 22122


@pytest.fixture
def client(default_host: Host) -> pymemcache.Client:
    return pymemcache.Client(default_host)


@pytest.fixture
def pooled_client(default_host: Host) -> pymemcache.PooledClient:
    return pymemcache.PooledClient(default_host)


@pytest.fixture
def flush_default_host(client: pymemcache.Client):
    client.flush_all()


@pytest.fixture
def flush_pooled_host(client: pymemcache.PooledClient):
    client.flush_all()


@pytest.mark.usefixtures("flush_default_host")
def test_fixed_window_with_elastic_expiry(client: pymemcache.Client):
    storage = MemcachedStorage(client)
    limiter = FixedWindowElasticExpiryRateLimiter(storage)
    limit = RateLimitItemPerSecond(10, 2)

    try:
        assert all([limiter.hit(limit) for _ in range(0, 10)]) is True

        time.sleep(1)
        assert limiter.hit(limit) is False

        time.sleep(1)
        assert limiter.hit(limit) is False
    except pymemcache.exceptions.MemcacheIllegalInputError as e:
        if str(e) == "cas must be integer, string, or bytes, got bad value: None":
            pytest.xfail(reason="Known flaky test due to logic error in MemcachedStorage.incr().")
        else:
            raise


@pytest.mark.usefixtures("flush_pooled_host")
def test_fixed_window_with_elastic_expiry_concurrency(pooled_client: pymemcache.PooledClient):
    storage = MemcachedStorage(pooled_client)
    limiter = FixedWindowElasticExpiryRateLimiter(storage)
    limit = RateLimitItemPerSecond(10, 2)

    start = time.time()

    def _c() -> None:
        for _ in range(0, 5):
            try:
                limiter.hit(limit)
            except Exception:
                traceback.print_exc()
                raise

    t1 = threading.Thread(target=_c)
    t2 = threading.Thread(target=_c)

    t1.start()
    t2.start()

    t1.join()
    t2.join()

    window_stats = limiter.get_window_stats(limit)
    assert window_stats.remaining_count == 0
    assert start + 2 <= window_stats.reset_time <= start + 3
    assert storage.get(limit.key_for()) == 10


@pytest.mark.usefixtures("flush_default_host")
def test_moving_window(client: pymemcache.Client):
    storage = MemcachedStorage(client)
    with pytest.raises(TypeError):
        # Ignore the type error here because that's exactly what we're testing for.
        MovingWindowRateLimiter(storage)  # type: ignore



import time

import pytest

from freiner import RateLimitItemPerMinute, RateLimitItemPerSecond
from freiner.storage.memory import MemoryStorage
from freiner.strategies.fixed_window import FixedWindowRateLimiter
from freiner.strategies.moving_window import MovingWindowRateLimiter

from ..util import freeze_time


@pytest.fixture
def storage() -> MemoryStorage:
    memory_storage = MemoryStorage()
    assert memory_storage.check() is True
    return memory_storage


def test_in_memory(storage: MemoryStorage):
    limiter = FixedWindowRateLimiter(storage)
    with freeze_time() as frozen_datetime:
        per_min = RateLimitItemPerMinute(10)

        for _ in range(0, 10):
            assert limiter.hit(per_min) is True
        assert limiter.hit(per_min) is False

        frozen_datetime.tick(60)
        assert limiter.hit(per_min) is True


def test_fixed_window_clear(storage: MemoryStorage):
    limiter = FixedWindowRateLimiter(storage)
    with freeze_time():
        per_min = RateLimitItemPerMinute(1)

        assert limiter.hit(per_min) is True
        assert limiter.hit(per_min) is False

        limiter.clear(per_min)
        assert limiter.hit(per_min) is True


def test_moving_window_clear(storage: MemoryStorage):
    limiter = MovingWindowRateLimiter(storage)
    with freeze_time():
        per_min = RateLimitItemPerMinute(1)

        assert limiter.hit(per_min) is True
        assert limiter.hit(per_min) is False

        limiter.clear(per_min)
        assert limiter.hit(per_min) is True


def test_reset(storage: MemoryStorage):
    limiter = FixedWindowRateLimiter(storage)
    with freeze_time():
        per_min = RateLimitItemPerMinute(10)

        for _ in range(0, 10):
            assert limiter.hit(per_min) is True
        assert limiter.hit(per_min) is False

        storage.reset()
        for _ in range(0, 10):
            assert limiter.hit(per_min) is True
        assert limiter.hit(per_min) is False


def test_expiry_fixed_window(storage: MemoryStorage):
    limiter = FixedWindowRateLimiter(storage)
    with freeze_time() as frozen_datetime:
        per_min = RateLimitItemPerMinute(10)
        per_sec = RateLimitItemPerSecond(1)

        for _ in range(0, 10):
            assert limiter.hit(per_min) is True
        assert limiter.hit(per_min) is False

        frozen_datetime.tick(60)
        # touch another key and yield
        assert limiter.hit(per_sec) is True
        time.sleep(0.02)
        assert per_min.key_for() not in storage.storage


def test_expiry_moving_window(storage: MemoryStorage):
    limiter = MovingWindowRateLimiter(storage)
    with freeze_time() as frozen_datetime:
        per_min = RateLimitItemPerMinute(10)
        per_sec = RateLimitItemPerSecond(1)

        for _ in range(0, 2):
            for _ in range(0, 10):
                assert limiter.hit(per_min) is True

            frozen_datetime.tick(60)
            # touch another key and yield
            assert limiter.hit(per_sec) is True
            time.sleep(0.02)
            assert storage.events[per_min.key_for()] == []



import functools
import math
import time
from pathlib import Path
from typing import Any, Callable, ContextManager

from freezegun import freeze_time as _freeze_time
from freezegun.api import FrozenDateTimeFactory


ROOTDIR = Path(__file__).parent.parent
DOCKERDIR = ROOTDIR / ".docker"


def fixed_start(func: Callable[..., Any]):
    @functools.wraps(func)
    def _inner(*args: Any, **kwargs: Any):
        start = time.time()
        while time.time() < math.ceil(start):
            time.sleep(0.01)
        return func(*args, **kwargs)

    return _inner


def freeze_time() -> ContextManager[FrozenDateTimeFactory]:
    f = _freeze_time()
    # This is a necessary fix for our testing, and the third-party type hints aren't sufficient
    f.ignore = tuple(set(f.ignore) - {"threading"})  # type: ignore
    return f



import re
from typing import Sequence, Type, cast

from .limits import GRANULARITIES, RateLimitItem


SEPARATORS = re.compile(r"[,;|]")
SINGLE_EXPR = re.compile(
    r"""
    \s*([0-9]+)
    \s*(?:/|\s*per\s*)
    \s*([0-9]+)
    *\s*(hour|minute|second|day|month|year)s?\s*""",
    re.IGNORECASE | re.VERBOSE,
)
EXPR = re.compile(
    r"^{SINGLE}(?:{SEPARATORS}{SINGLE})*$".format(
        SINGLE=SINGLE_EXPR.pattern, SEPARATORS=SEPARATORS.pattern
    ),
    re.IGNORECASE | re.VERBOSE,
)


def parse_many(limit_string: str) -> Sequence[RateLimitItem]:
    """
    Parses rate limits in string notation containing multiple rate limits
    (e.g. '1/second; 5/minute').

    :param limit_string: The rate limit string. using :ref:`ratelimit-string`.
    :raises TypeError: If something other than a string was supplied.
    :raises ValueError: If the string notation is invalid.
    :return: A sequence of :class:`freiner.limits.RateLimitItem` instances.
    """

    if not isinstance(limit_string, str):
        raise TypeError("Invalid rate limit string supplied.")

    if not EXPR.match(limit_string):
        raise ValueError(f"Failed to parse rate limit string: {limit_string}")

    limits = []
    for limit in SEPARATORS.split(limit_string):
        # This cast is fine because we already verified that it will match
        # in the EXPR.match check above.
        limit_match = cast(re.Match, SINGLE_EXPR.match(limit))
        amount, multiples, granularity_string = limit_match.groups()
        granularity = granularity_from_string(granularity_string)
        limits.append(granularity(amount, multiples))

    return tuple(limits)


def parse(limit_string: str) -> RateLimitItem:
    """
    Parses a single rate limit in string notation (e.g. '1/second' or '1 per second').

    # noqa: DAR402

    :param limit_string: The rate limit string. using :ref:`ratelimit-string`.
    :raises TypeError: If something other than a string was supplied.
    :raises ValueError: If the string notation is invalid.
    :return: An instance of :class:`freiner.limits.RateLimitItem`.
    """

    return parse_many(limit_string)[0]


def granularity_from_string(granularity_string: str) -> Type[RateLimitItem]:
    """
    :param granularity_string: A representation of the granularity (eg. "second", "minute").
    :raises ValueError: If the supplied granularity is unknown.
    :return: A subclass of :class:`freiner.limits.RateLimitItem`.
    """

    for granularity in GRANULARITIES.values():
        if granularity.check_granularity_string(granularity_string):
            return granularity

    raise ValueError(f"No granularity matched for: {granularity_string}")



from typing import Any, NamedTuple

from typing_extensions import Protocol, runtime_checkable

from freiner.limits import RateLimitItem


class WindowStats(NamedTuple):
    reset_time: float
    remaining_count: int


@runtime_checkable
class RateLimiter(Protocol):
    def hit(self, item: RateLimitItem, *identifiers: Any) -> bool:
        """
        Creates a hit on the rate limit and returns ``True`` if successful.

        # noqa: DAR202

        :param item: A :class:`freiner.limits.RateLimitItem` instance.
        :param identifiers: A variable list of stringable objects to uniquely identify the limit.
        :return: ``True`` if the request was successful, or ``False`` if the rate limit had been exceeded.
        """

    def test(self, item: RateLimitItem, *identifiers: Any) -> bool:
        """
        Checks the rate limit and returns ``True`` if it is not currently exceeded.

        # noqa: DAR202

        :param item: A :class:`freiner.limits.RateLimitItem` instance.
        :param identifiers: A variable list of stringable objects to uniquely identify the limit.
        :return: ``True`` if the rate limit has not yet been exceeded, or ``False`` if it has.
        """

    def get_window_stats(self, item: RateLimitItem, *identifiers: Any) -> WindowStats:
        """
        Retrieve statistics about the number of requests remaining within the given limit.

        # noqa: DAR202

        :param item: A :class:`freiner.limits.RateLimitItem` instance.
        :param identifiers: A variable list of stringable objects to uniquely identify the limit.
        :return: tuple (reset time (float), remaining (int))
        """

    def clear(self, item: RateLimitItem, *identifiers: Any) -> None:
        """
        Resets the request counter for a given limit to zero.

        :param item: a :class:`freiner.limits.RateLimitItem` instance
        :param identifiers: A variable list of stringable objects to uniquely identify the limit.
        """


__all__ = [
    "RateLimiter",
    "WindowStats",
]



import threading
import time
from collections import Counter
from typing import Counter as CounterType
from typing import Dict, List, Optional

from . import MovingWindow


class _LockableEntry:
    __slots__ = ("atime", "expiry", "_lock")

    def __init__(self, expiry: float):
        self.atime: float = time.time()
        self.expiry: float = self.atime + expiry

        self._lock = threading.RLock()

    def acquire(self) -> None:
        self._lock.acquire()

    def release(self) -> None:
        self._lock.release()

    def __enter__(self):
        self.acquire()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.release()

    def __repr__(self) -> str:
        return f"MemoryLockableEntry<atime={self.atime}, expiry={self.expiry}>"  # pragma: no cover


class MemoryStorage:
    """
    rate limit storage using :py:class:`collections.Counter`
    as an in memory storage for fixed and elastic window strategies,
    and a simple list to implement moving window strategy.
    """

    def __init__(self) -> None:
        self.storage: CounterType[str] = Counter()
        self.expirations: Dict[str, float] = {}
        self.events: Dict[str, List[_LockableEntry]] = {}

        self.timer = threading.Timer(0.01, self.__expire_events)
        self.timer.start()

    def __expire_events(self) -> None:
        for key in list(self.events.keys()):
            for event in list(self.events[key]):
                with event:
                    if event.expiry <= time.time() and event in self.events[key]:
                        self.events[key].remove(event)

        for key in list(self.expirations.keys()):
            if self.expirations[key] <= time.time():
                self.storage.pop(key, None)
                self.expirations.pop(key, None)

    def __schedule_expiry(self) -> None:
        if not self.timer.is_alive():
            self.timer = threading.Timer(0.01, self.__expire_events)
            self.timer.start()

    def incr(self, key: str, expiry: int, elastic_expiry: bool = False) -> int:
        """
        Increments the counter for the given rate limit key.

        :param key: The key to increment.
        :param expiry: Amount in seconds for the key to expire in.
        :param elastic_expiry: Whether to keep extending the rate limit window every hit.
        :return: The number of hits currently on the rate limit for the given key.
        """

        self.get(key)
        self.__schedule_expiry()
        self.storage[key] += 1
        if elastic_expiry or self.storage[key] == 1:
            self.expirations[key] = time.time() + expiry
        return self.storage.get(key, 0)

    def get(self, key: str) -> int:
        """
        Retrieve the current request count for the given rate limit key.

        :param key: The key to get the counter value for.
        """

        if self.expirations.get(key, 0) <= time.time():
            self.storage.pop(key, None)
            self.expirations.pop(key, None)
        return self.storage.get(key, 0)

    def clear(self, key: str) -> None:
        """
        Resets the rate limit for the given key.

        :param key: The key to clear rate limits for.
        """
        self.storage.pop(key, None)
        self.expirations.pop(key, None)
        self.events.pop(key, None)

    def acquire_entry(self, key: str, limit: int, expiry: int, no_add: bool = False) -> bool:
        """
        :param key: The rate limit key to acquire an entry in.
        :param limit: The total amount of entries allowed before hitting the rate limit.
        :param expiry: Amount in seconds for the acquired entry to expire in.
        :param no_add: If False, an entry is not actually acquired but instead serves as a 'check'.
        """

        self.events.setdefault(key, [])
        self.__schedule_expiry()
        timestamp = time.time()
        entry: Optional[_LockableEntry]
        try:
            entry = self.events[key][limit - 1]
        except IndexError:
            entry = None

        if entry and entry.atime > timestamp - expiry:
            return False
        else:
            if not no_add:
                self.events[key].insert(0, _LockableEntry(expiry))
            return True

    def get_expiry(self, key: str) -> float:
        """
        Retrieve the expected expiry time for the given rate limit key.

        :param key: The key to get the expiry time for.
        :return: The time at which the current rate limit for the given key ends.
        """

        return self.expirations.get(key, -1)

    def get_num_acquired(self, key: str, expiry: int) -> int:
        """
        returns the number of entries already acquired

        :param key: rate limit key to acquire an entry in
        :param expiry: expiry of the entry
        """

        timestamp = time.time()
        if self.events.get(key):
            return len([k for k in self.events[key] if k.atime > timestamp - expiry])
        else:
            return 0

    def get_moving_window(self, key: str, limit: int, expiry: int) -> MovingWindow:
        """
        Retrieves the starting point and the number of entries in the moving window.

        :param key: The rate limit key to retrieve statistics about.
        :param limit: The total amount of entries allowed before hitting the rate limit.
        :param expiry: Amount in seconds for the acquired entry to expire in.
        :return: (start of window, number of acquired entries)
        """

        timestamp = time.time()
        acquired = self.get_num_acquired(key, expiry)
        for item in self.events.get(key, []):
            if item.atime > timestamp - expiry:
                return MovingWindow(item.atime, acquired)
        return MovingWindow(timestamp, acquired)

    def check(self) -> bool:
        """
        Check if the connection to the storage backend is healthy.
        """

        return True

    def reset(self) -> None:
        self.storage.clear()
        self.expirations.clear()
        self.events.clear()


__all__ = [
    "MemoryStorage",
]



import time

import pytest

from freiner.limits import RateLimitItemPerMinute, RateLimitItemPerSecond
from freiner.storage.memory import MemoryStorage
from freiner.strategies.fixed_window import FixedWindowRateLimiter
from freiner.strategies.fixed_window_elastic import FixedWindowElasticExpiryRateLimiter
from freiner.strategies.moving_window import MovingWindowRateLimiter

from ..util import freeze_time


@pytest.fixture
def storage() -> MemoryStorage:
    return MemoryStorage()


def test_fixed_window_simple(storage: MemoryStorage):
    limiter = FixedWindowRateLimiter(storage)
    with freeze_time():
        limit = RateLimitItemPerSecond(2, 1)

        assert limiter.test(limit) is True
        assert limiter.hit(limit) is True
        assert limiter.test(limit) is True
        assert limiter.hit(limit) is True
        assert limiter.test(limit) is False
        assert limiter.hit(limit) is False


def test_fixed_window(storage: MemoryStorage):
    limiter = FixedWindowRateLimiter(storage)
    with freeze_time() as frozen_datetime:
        limit = RateLimitItemPerSecond(10, 2)
        start = time.time()

        assert all([limiter.hit(limit) for _ in range(0, 10)]) is True
        assert limiter.hit(limit) is False

        frozen_datetime.tick(1)
        assert limiter.hit(limit) is False
        window_stats = limiter.get_window_stats(limit)
        assert window_stats.reset_time == start + 2
        assert window_stats.remaining_count == 0

        frozen_datetime.tick(1)
        assert limiter.get_window_stats(limit).remaining_count == 10
        assert limiter.hit(limit) is True


def test_fixed_window_with_elastic_expiry(storage: MemoryStorage):
    limiter = FixedWindowElasticExpiryRateLimiter(storage)
    with freeze_time() as frozen_datetime:
        limit = RateLimitItemPerSecond(10, 2)
        start = time.time()

        assert all([limiter.hit(limit) for _ in range(0, 10)]) is True
        assert limiter.hit(limit) is False

        frozen_datetime.tick(1)
        assert limiter.hit(limit) is False
        window_stats = limiter.get_window_stats(limit)
        # three extensions to the expiry
        assert window_stats.reset_time == start + 3
        assert window_stats.remaining_count == 0

        frozen_datetime.tick(1)
        assert limiter.hit(limit) is False

        frozen_datetime.tick(3)
        start = time.time()
        assert limiter.hit(limit) is True
        window_stats = limiter.get_window_stats(limit)
        assert window_stats.reset_time == start + 2
        assert window_stats.remaining_count == 9


def test_moving_window_simple(storage: MemoryStorage):
    limiter = MovingWindowRateLimiter(storage)
    with freeze_time():
        limit = RateLimitItemPerSecond(2, 1)

        assert limiter.test(limit) is True
        assert limiter.hit(limit) is True
        assert limiter.test(limit) is True
        assert limiter.hit(limit) is True
        assert limiter.test(limit) is False
        assert limiter.hit(limit) is False


def test_moving_window(storage: MemoryStorage):
    limiter = MovingWindowRateLimiter(storage)
    with freeze_time() as frozen_datetime:
        limit = RateLimitItemPerMinute(10)

        for i in range(0, 5):
            assert limiter.hit(limit) is True
            assert limiter.hit(limit) is True
            assert limiter.get_window_stats(limit).remaining_count == 10 - ((i + 1) * 2)
            frozen_datetime.tick(10)

        assert limiter.get_window_stats(limit).remaining_count == 0
        assert limiter.hit(limit) is False

        frozen_datetime.tick(20)
        window_stats = limiter.get_window_stats(limit)
        assert window_stats.reset_time == time.time() + 30
        assert window_stats.remaining_count == 4

        frozen_datetime.tick(30)
        assert limiter.get_window_stats(limit).remaining_count == 10



import time

import pytest
from redis.sentinel import Sentinel

from freiner.limits import RateLimitItemPerSecond
from freiner.storage.redis_sentinel import RedisSentinelStorage
from freiner.strategies.fixed_window_elastic import FixedWindowElasticExpiryRateLimiter


@pytest.fixture
def client() -> Sentinel:
    return Sentinel([("localhost", 26379)])


@pytest.fixture
def service_name() -> str:
    return "localhost-redis-sentinel"


@pytest.fixture
def storage(client: Sentinel, service_name: str) -> RedisSentinelStorage:
    return RedisSentinelStorage(client, service_name)


@pytest.fixture(autouse=True)
def flush_client(client: Sentinel, service_name: str):
    client.master_for(service_name).flushall()


def test_fixed_window_with_elastic_expiry(storage: RedisSentinelStorage):
    limiter = FixedWindowElasticExpiryRateLimiter(storage)
    limit = RateLimitItemPerSecond(10, 2)

    assert all([limiter.hit(limit) for _ in range(0, 10)]) is True

    time.sleep(1)
    assert limiter.hit(limit) is False

    time.sleep(1)
    assert limiter.hit(limit) is False
    assert limiter.get_window_stats(limit).remaining_count == 0



from .limits import (
    RateLimitItem,
    RateLimitItemPerDay,
    RateLimitItemPerHour,
    RateLimitItemPerMinute,
    RateLimitItemPerMonth,
    RateLimitItemPerSecond,
    RateLimitItemPerYear,
)
from .storage import FixedWindowStorage, MovingWindow, MovingWindowStorage
from .storage.memory import MemoryStorage
from .strategies import RateLimiter, WindowStats
from .strategies.fixed_window import FixedWindowRateLimiter
from .strategies.fixed_window_elastic import FixedWindowElasticExpiryRateLimiter
from .strategies.moving_window import MovingWindowRateLimiter
from .util import parse, parse_many


__version__ = "3.0.4"

__all__ = [
    "FixedWindowStorage",
    "MovingWindow",
    "MovingWindowStorage",
    "MemoryStorage",
    "RateLimitItem",
    "RateLimitItemPerYear",
    "RateLimitItemPerMonth",
    "RateLimitItemPerDay",
    "RateLimitItemPerHour",
    "RateLimitItemPerMinute",
    "RateLimitItemPerSecond",
    "RateLimiter",
    "WindowStats",
    "FixedWindowRateLimiter",
    "FixedWindowElasticExpiryRateLimiter",
    "MovingWindowRateLimiter",
    "parse",
    "parse_many",
]



import time
from typing import Any, Callable, Tuple, cast

import redis

from . import MovingWindow


class RedisInteractor:
    SCRIPT_MOVING_WINDOW = """
        local items = redis.call('lrange', KEYS[1], 0, tonumber(ARGV[2]))
        local expiry = tonumber(ARGV[1])
        local a = 0
        local oldest = nil
        for idx=1,#items do
            if tonumber(items[idx]) > expiry then
                a = a + 1
                if oldest == nil then
                    oldest = tonumber(items[idx])
                end
            else
                break
            end
        end
        return {oldest, a}
        """

    SCRIPT_ACQUIRE_MOVING_WINDOW = """
        local entry = redis.call('lindex', KEYS[1], tonumber(ARGV[2]) - 1)
        local timestamp = tonumber(ARGV[1])
        local expiry = tonumber(ARGV[3])
        if entry and tonumber(entry) > timestamp - expiry then
            return false
        end
        local limit = tonumber(ARGV[2])
        local no_add = tonumber(ARGV[4])
        if no_add == 0 then
            redis.call('lpush', KEYS[1], timestamp)
            redis.call('ltrim', KEYS[1], 0, limit - 1)
            redis.call('expire', KEYS[1], expiry)
        end
        return true
        """

    SCRIPT_CLEAR_KEYS = """
        local keys = redis.call('keys', KEYS[1])
        local res = 0
        for i=1,#keys,5000 do
            res = res + redis.call(
                'del', unpack(keys, i, math.min(i+4999, #keys))
            )
        end
        return res
        """

    SCRIPT_INCR_EXPIRE = """
        local current
        current = redis.call("incr", KEYS[1])
        if tonumber(current) == 1 then
            redis.call("expire", KEYS[1], ARGV[1])
        end
        return current
    """

    def initialize_storage(self, connection: redis.Redis):
        moving_window_script = connection.register_script(self.SCRIPT_MOVING_WINDOW)
        self.lua_moving_window = cast(
            Callable[[Tuple[str], Tuple[float, int]], Tuple[float, int]],
            moving_window_script,
        )

        acquire_window_script = connection.register_script(self.SCRIPT_ACQUIRE_MOVING_WINDOW)
        self.lua_acquire_window = cast(
            Callable[[Tuple[str], Tuple[float, int, int, int]], bool],
            acquire_window_script,
        )

        clear_keys_script = connection.register_script(self.SCRIPT_CLEAR_KEYS)
        self.lua_clear_keys = cast(
            Callable[[Tuple[str]], int],
            clear_keys_script,
        )

        incr_expire_script = connection.register_script(RedisStorage.SCRIPT_INCR_EXPIRE)
        self.lua_incr_expire = cast(
            Callable[[Tuple[str], Tuple[int]], int],
            incr_expire_script,
        )

    def get_moving_window(self, key: str, limit: int, expiry: int) -> MovingWindow:
        """
        Retrieves the starting point and the number of entries in the moving window.

        :param key: The rate limit key to retrieve statistics about.
        :param limit: The total amount of entries allowed before hitting the rate limit.
        :param expiry: Amount in seconds for the acquired entry to expire in.
        :return: (start of window, number of acquired entries)
        """

        timestamp = time.time()
        window = self.lua_moving_window((key,), (timestamp - expiry, limit))
        return MovingWindow(window[0], window[1])

    def _incr(self, key: str, expiry: int, connection: redis.Redis, elastic_expiry: bool = False):
        """
        Increments the counter for the given rate limit key.

        :param key: The key to increment.
        :param expiry: Amount in seconds for the key to expire in.
        :param connection: Redis connection.
        :param elastic_expiry: Whether to keep extending the rate limit window every hit.
        :return: The number of hits currently on the rate limit for the given key.
        """

        value = connection.incr(key)
        if elastic_expiry or value == 1:
            connection.expire(key, expiry)
        return value

    def _get(self, key: str, connection: redis.Redis) -> int:
        """
        Retrieve the current request count for the given rate limit key.

        :param key: The key to get the counter value for.
        :param connection: Redis connection.
        """

        return int(connection.get(key) or 0)

    def _clear(self, key: str, connection: redis.Redis) -> None:
        """
        Resets the rate limit for the given key.

        :param key: The key to clear rate limits for.
        :param connection: Redis connection.
        """

        connection.delete(key)

    def _reset(self) -> int:
        return self.lua_clear_keys(("LIMITER*",))

    def _acquire_entry(
        self, key: str, limit: int, expiry: int, connection: redis.Redis, no_add: bool = False
    ) -> bool:
        """
        :param key: The rate limit key to acquire an entry in.
        :param limit: The total amount of entries allowed before hitting the rate limit.
        :param expiry: Amount in seconds for the acquired entry to expire in.
        :param connection: Redis connection.
        :param no_add: If False, an entry is not actually acquired but instead serves as a 'check'.
        """

        timestamp = time.time()
        acquired = self.lua_acquire_window(
            (key,),
            (timestamp, limit, expiry, int(no_add)),
        )
        return bool(acquired)

    def _get_expiry(self, key: str, connection: redis.Redis) -> float:
        """
        Retrieve the expected expiry time for the given rate limit key.

        :param key: The key to get the expiry time for.
        :param connection: Redis connection.
        :return: The time at which the current rate limit for the given key ends.
        """

        return max(connection.ttl(key), 0) + time.time()

    def _check(self, connection: redis.Redis) -> bool:
        """
        Check if the connection to the storage backend is healthy.

        :param connection: Redis connection.
        """

        try:
            return connection.ping()
        except:  # noqa
            return False


class RedisStorage(RedisInteractor):
    """
    Rate limit storage with redis as backend.

    Depends on the `redis` library.
    """

    def __init__(self, client: redis.Redis) -> None:
        self._client = client
        self.initialize_storage(self._client)

    @classmethod
    def from_uri(cls, uri: str, **options: Any) -> "RedisStorage":
        """
        :param uri: URI of the form `redis://[:password]@host:port`, `redis://[:password]@host:port/db`,
                    `rediss://[:password]@host:port`, `unix:///path/to/sock` etc. This uri is passed
                    directly to :func:`redis.from_url`.
        :param options: All remaining keyword arguments are passed directly to the constructor
                        of :class:`redis.Redis`.
        """

        client: redis.Redis = redis.from_url(uri, **options)
        return cls(client)

    def incr(self, key: str, expiry: int, elastic_expiry: bool = False) -> int:
        """
        Increments the counter for the given rate limit key.

        :param key: The key to increment.
        :param expiry: Amount in seconds for the key to expire in.
        :param elastic_expiry: Whether to keep extending the rate limit window every hit.
        :return: The number of hits currently on the rate limit for the given key.
        """

        if elastic_expiry:
            return self._incr(key, expiry, self._client, elastic_expiry)
        else:
            return self.lua_incr_expire((key,), (expiry,))

    def get(self, key: str) -> int:
        """
        Retrieve the current request count for the given rate limit key.

        :param key: The key to get the counter value for.
        """

        return self._get(key, self._client)

    def clear(self, key: str) -> None:
        """
        Resets the rate limit for the given key.

        :param key: The key to clear rate limits for.
        """

        self._clear(key, self._client)

    def acquire_entry(self, key: str, limit: int, expiry: int, no_add: bool = False) -> bool:
        """
        :param key: The rate limit key to acquire an entry in.
        :param limit: The total amount of entries allowed before hitting the rate limit.
        :param expiry: Amount in seconds for the acquired entry to expire in.
        :param no_add: If False, an entry is not actually acquired but instead serves as a 'check'.
        """

        return self._acquire_entry(key, limit, expiry, self._client, no_add=no_add)

    def get_expiry(self, key: str) -> float:
        """
        Retrieve the expected expiry time for the given rate limit key.

        :param key: The key to get the expiry time for.
        :return: The time at which the current rate limit for the given key ends.
        """

        return self._get_expiry(key, self._client)

    def check(self) -> bool:
        """
        Check if the connection to the storage backend is healthy.
        """

        return self._check(self._client)

    def reset(self) -> None:
        """
        This function calls a Lua Script to delete keys prefixed with 'LIMITER'
        in blocks of 5000.

        .. warning::
           This operation was designed to be fast, but was not tested
           on a large production based system. Be careful with its usage as it
           could be slow on very large data sets.
        """

        self._reset()


__all__ = [
    "RedisInteractor",
    "RedisStorage",
]



import time

import pytest
import redis

from freiner.limits import RateLimitItemPerSecond
from freiner.storage.redis import RedisStorage
from freiner.strategies.fixed_window_elastic import FixedWindowElasticExpiryRateLimiter
from freiner.strategies.moving_window import MovingWindowRateLimiter


@pytest.fixture
def client() -> redis.Redis:
    return redis.from_url("redis://localhost:7379")


@pytest.fixture
def storage(client: redis.Redis) -> RedisStorage:
    return RedisStorage(client)


@pytest.fixture(autouse=True)
def flush_client(client: redis.Redis):
    client.flushall()


def test_fixed_window_with_elastic_expiry(storage: RedisStorage):
    limiter = FixedWindowElasticExpiryRateLimiter(storage)
    limit = RateLimitItemPerSecond(10, 2)

    assert all([limiter.hit(limit) for _ in range(0, 10)]) is True

    time.sleep(1)
    assert limiter.hit(limit) is False

    time.sleep(1)
    assert limiter.hit(limit) is False
    assert limiter.get_window_stats(limit).remaining_count == 0


def test_moving_window(storage: RedisStorage):
    limiter = MovingWindowRateLimiter(storage)
    limit = RateLimitItemPerSecond(10, 2)

    for i in range(0, 10):
        assert limiter.hit(limit) is True
        assert limiter.get_window_stats(limit).remaining_count == 10 - (i + 1)
        time.sleep(2 * 0.095)

    assert limiter.hit(limit) is False

    time.sleep(0.4)
    assert limiter.hit(limit) is True
    assert limiter.hit(limit) is True
    assert limiter.get_window_stats(limit).remaining_count == 0
